// BenLincoln.DigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.DigitalAudio.

// BenLincoln.DigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.DigitalAudio
{
    public class WAVFile : IDisposable, BLS.IProgress
    {
        // descriptor attributes, in order of their appearance in the header
        protected uint _ChunkID = 0x52494646;       // will always be 0x52494646 ("RIFF" in ASCII) in a WAV file
        protected uint _ChunkSize;                  // size of the WAV data in bytes + 36 bytes for the rest of the header after this point
        protected uint _Format = 0x57415645;        // will always be 0x57415645 ("WAVE" in ASCII) in a WAV file

        protected uint _SubChunk1ID = 0x666d7420;   // will always be 0x666d7420 ("fmt " in ASCII) in a WAV file
        protected uint _SubChunk1Size = 16;         // 16 = PCM linear quantization, 16 more bytes in this part of the header, compressed formats not supported by this class
        protected ushort _AudioFormat = 1;          // 1 = PCM linear quantization, compressed formats not supported by this class
        protected ushort _NumChannels;              // 1 = mono, 2 = stereo, etc.
        protected uint _SampleRate;                 // samples per second per channel
        protected uint _ByteRate;                   // SampleRate * NumChannels * (BitsPerSample / 8)
        protected ushort _BlockAlign;               // NumChannels * (BitsPerSample / 8), that is the number of bytes for one sample of all channels
        protected ushort _BitsPerSample;            // 8, 16, etc.

        protected uint _SubChunk2ID = 0x64617461;   // will always be 0x64617461 ("data" in ASCII) in a WAV file
        protected uint _SubChunk2Size;              // NumSamples * NumChannels * (BitsPerSample / 8), size of the data portion in bytes

        protected int _HeaderLength = 44;           // length of the header in bytes

        // descriptor attributes that aren't in the header
        protected uint _NumSamplesPerChannel;       // total number of samples in the file for each channel

        // a temp file to hold the data
        protected string _TempFilePath;

        // IO progress - can do read/write this way
        protected double _IOPercent = 0;

        #region Properties

        public uint ChunkID
        {
            get
            {
                return _ChunkID;
            }
        }

        public uint ChunkSize
        {
            get
            {
                return _ChunkSize;
            }
        }

        public uint Format
        {
            get
            {
                return _Format;
            }
        }

        public uint SubChunk1ID
        {
            get
            {
                return _SubChunk1ID;
            }
        }

        public uint SubChunk1Size
        {
            get
            {
                return _SubChunk1Size;
            }
        }

        public ushort AudioFormat
        {
            get
            {
                return _AudioFormat;
            }
        }

        public ushort NumChannels
        {
            get
            {
                return _NumChannels;
            }
        }

        public uint SampleRate
        {
            get
            {
                return _SampleRate;
            }
        }

        public uint ByteRate
        {
            get
            {
                return _ByteRate;
            }
        }

        public ushort BlockAlign
        {
            get
            {
                return _BlockAlign;
            }
        }

        public ushort BitsPerSample
        {
            get
            {
                return _BitsPerSample;
            }
        }

        public uint SubChunk2ID
        {
            get
            {
                return _SubChunk2ID;
            }
        }

        public uint SubChunk2Size
        {
            get
            {
                return _SubChunk2Size;
            }
        }

        public int HeaderLength
        {
            get
            {
                return _HeaderLength;
            }
        }

        public uint NumSamplesPerChannel
        {
            get
            {
                return _NumSamplesPerChannel;
            }
        }

        public string TempFilePath
        {
            get
            {
                return _TempFilePath;
            }
            set
            {
                _TempFilePath = value;
            }
        }

        #endregion

        public WAVFile(ushort numChannels, uint sampleRate, ushort bitsPerSample, uint numSamplesPerChannel)
        {
            // validate passed values
            if ((numChannels < 1) || (numChannels > 10))
            {
                throw new BLS.InvalidParameterValueException("Number of channels of digital audio must be between 1 and 10.");
            }
            if ((bitsPerSample != 8) && (bitsPerSample != 16))
            {
                throw new BLS.InvalidParameterValueException("Bits per sample must be 8 or 16");
            }

            _NumChannels = numChannels;
            _SampleRate = sampleRate;
            _BitsPerSample = bitsPerSample;
            _NumSamplesPerChannel = numSamplesPerChannel;

            // determine calculated variables
            _ByteRate = (uint)(sampleRate * numChannels * (bitsPerSample / 8));
            _BlockAlign = (ushort)(numChannels * (bitsPerSample / 8));
            _SubChunk2Size = (uint)(numSamplesPerChannel * numChannels * (bitsPerSample / 8));
            _ChunkSize = (uint)(_SubChunk2Size + 36);
            //data must be an even number of bytes long
            if (_ChunkSize % 2 == 1)
            {
                _ChunkSize++;
            }
            string tPath = Path.GetTempFileName();
            _TempFilePath = tPath + ".wav";
            // delete the non-.wav extension temp file
            File.Delete(tPath);
            WriteHeader(_TempFilePath);
        }

        public static BLDA.WAVFile FromStreamHeaderOnly(Stream inStream)
        {
            BinaryReader iReader = new BinaryReader(inStream);
            uint chunkID = iReader.ReadUInt32();
            if (chunkID != 0x46464952)
            {
                throw new BLS.InvalidFileFormatException("Data format unrecognized. Stream does not begin with 'RIFF' header.");
            }

            uint chunkSize = iReader.ReadUInt32();

            uint format = iReader.ReadUInt32();
            if (format != 0x45564157)
            {
                throw new BLS.InvalidFileFormatException("Data format unrecognized. RIFF format ID is not 'WAVE'.");
            }

            uint subChunk1ID = iReader.ReadUInt32();
            if (subChunk1ID != 0x20746d66)
            {
                throw new BLS.InvalidFileFormatException("Data format unrecognized. Sub-chunk 1 ID is not 'fmt '.");
            }

            uint subChunk1Size = iReader.ReadUInt32();
            ushort audioFormat = iReader.ReadUInt16();
            if ((subChunk1Size != 16) || (audioFormat != 1))
            {
                throw new BLS.InvalidFileFormatException("Data format unrecognized. Only uncompressed PCM linear encoding is supported.");
            }

            ushort numChannels = iReader.ReadUInt16();
            uint sampleRate = iReader.ReadUInt32();
            uint byteRate = iReader.ReadUInt32();
            ushort blockAlign = iReader.ReadUInt16();
            ushort bitsPerSample = iReader.ReadUInt16();

            uint subChunk2ID = iReader.ReadUInt32();
            if (subChunk2ID != 0x61746164)
            {
                throw new BLS.InvalidFileFormatException("Data format unrecognized. Sub-chunk 2 ID is not 'data'.");
            }

            uint subChunk2Size = iReader.ReadUInt32();

            // try to create a returnable WAVFile
            BLDA.WAVFile retWav;
            if (bitsPerSample == 8)
            {
                retWav = new BLDA.WAVFile(numChannels, sampleRate, bitsPerSample, (subChunk2Size / numChannels));
            }
            else if (bitsPerSample == 16)
            {
                retWav = new BLDA.WAVFile(numChannels, sampleRate, bitsPerSample, ((subChunk2Size / 2) / numChannels));
            }
            else
            {
                throw new BLS.InvalidFileFormatException("This stream indicates " + bitsPerSample.ToString() + " bits per sample. Only 8- and 16-bit formats are supported.");
            }

            //make sure the calculated values match the ones from the header
            //if (retWav.ChunkSize != chunkSize)
            //{
            //    throw new BLS.UnexpectedParameterValueException("ChunkSize mismatch. Header: " + chunkSize.ToString() +
            //        ", Calculated: " + retWav.ChunkSize.ToString());
            //}
            if (retWav.ByteRate != byteRate)
            {
                throw new BLS.UnexpectedParameterValueException("ByteRate mismatch. Header: " + byteRate.ToString() +
                    ", Calculated: " + retWav.ByteRate.ToString());
            }
            if (retWav.BlockAlign != blockAlign)
            {
                throw new BLS.UnexpectedParameterValueException("BlockAlign mismatch. Header: " + blockAlign.ToString() +
                    ", Calculated: " + retWav.BlockAlign.ToString());
            }
            if (retWav.SubChunk2Size != subChunk2Size)
            {
                throw new BLS.UnexpectedParameterValueException("SubChunk2Size mismatch. Header: " + subChunk2Size.ToString() +
                    ", Calculated: " + retWav.SubChunk2Size.ToString());
            }

            return retWav;
        }

        public static BLDA.WAVFile FromStreamFull(Stream inStream)
        {
            BLDA.WAVFile retWav = FromStreamHeaderOnly(inStream);

            //read the raw data into the temp file
            retWav.ReadData(inStream);
            inStream.Close();
            return retWav;
        }

        public void WriteHeader(string filePath)
        {
            FileStream outStream = new FileStream(filePath, FileMode.Append, FileAccess.Write);
            WriteHeader(outStream);
            outStream.Close();
        }

        public void WriteHeader(Stream outStream)
        {
            WriteHeader(outStream, true);
        }

        public void WriteHeader(Stream outStream, bool closeStream)
        {
            BinaryWriter outWriter = new BinaryWriter(outStream);
            //outWriter.Write(_ChunkID);
            WriteBigEndian(outWriter, _ChunkID);
            outWriter.Write(_ChunkSize);
            //outWriter.Write(_Format);
            WriteBigEndian(outWriter, _Format);
            //outWriter.Write(_SubChunk1ID);
            WriteBigEndian(outWriter, _SubChunk1ID);
            outWriter.Write(_SubChunk1Size);
            outWriter.Write(_AudioFormat);
            outWriter.Write(_NumChannels);
            outWriter.Write(_SampleRate);
            outWriter.Write(_ByteRate);
            outWriter.Write(_BlockAlign);
            outWriter.Write(_BitsPerSample);
            //outWriter.Write(_SubChunk2ID);
            WriteBigEndian(outWriter, _SubChunk2ID);
            outWriter.Write(_SubChunk2Size);
            if (closeStream)
            {
                outWriter.Close();
            }
        }

        public void WriteData(Stream outStream)
        {
            _IOPercent = 0;
            FileStream inStream = new FileStream(_TempFilePath, FileMode.Open, FileAccess.Read);

            long streamPosition = 0;
            long streamLength = inStream.Length;

            bool atEndOfStream = false;
            do
            {
                int currentByte = inStream.ReadByte();
                if (currentByte == -1)
                {
                    atEndOfStream = true;
                }
                else
                {
                    outStream.WriteByte((byte)currentByte);
                    _IOPercent = (double)streamPosition / (double)streamLength;
                    streamPosition++;
                }
            } while (!atEndOfStream);
            inStream.Close();
            _IOPercent = 1;
        }

        public void WriteDataThreaded(object outStream)
        {
            try
            {
                WriteData((Stream)outStream);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a Stream");
            }
        }

        public void ReadData(Stream inStream)
        {
            _IOPercent = 0;
            FileStream outStream = new FileStream(_TempFilePath, FileMode.Append, FileAccess.Write);

            long streamPosition = 0;
            long streamLength = 0;

            //// get the length of the input stream to support progress if possible
            //try
            //{
            //    streamLength = inStream.Length;
            //}
            //catch (NotSupportedException nEx)
            //{
            //    // calculate the expected number
            //    streamLength = (_NumChannels * _NumSamplesPerChannel) * (_BitsPerSample / 8);
            //}

            // use calculated input stream length to avoid reading more data than should be
            streamLength = (_NumChannels * _NumSamplesPerChannel) * (_BitsPerSample / 8);

            bool atEndOfStream = false;
            do
            {
                int currentByte = inStream.ReadByte();
                if ((currentByte == -1) || (streamPosition >= streamLength))
                {
                    atEndOfStream = true;
                }
                else
                {
                    outStream.WriteByte((byte)currentByte);
                    _IOPercent = (double)streamPosition / (double)streamLength;
                    streamPosition++;
                }
            } while (!atEndOfStream);
            outStream.Close();
            _IOPercent = 1;
        }

        public void ReadDataThreaded(object inStream)
        {
            try
            {
                ReadData((Stream)inStream);
            }
            catch (InvalidCastException ex)
            {
                throw new InvalidCastException("Passed object must be a Stream");
            }
        }

        protected void WriteBigEndian(BinaryWriter outWriter, uint littleEndianUInt)
        {
            outWriter.Write(GetNthByte(littleEndianUInt, 0));
            outWriter.Write(GetNthByte(littleEndianUInt, 1));
            outWriter.Write(GetNthByte(littleEndianUInt, 2));
            outWriter.Write(GetNthByte(littleEndianUInt, 3));
        }

        protected byte GetNthByte(uint inUInt, int byteNumber)
        {
            uint byteVal = (uint)(inUInt << (8 * byteNumber));
            byteVal >>= 24;
            return (byte)byteVal;
        }

        protected byte GetNthByte(ushort inUShort, int byteNumber)
        {
            uint byteVal = (uint)(inUShort << (8 * byteNumber));
            byteVal >>= 24;
            return (byte)byteVal;
        }

        protected void UpdateIOPercent(double newPercent)
        {
            _IOPercent = newPercent;
        }

        public double Progress()
        {
            return _IOPercent;
        }

        public void Dispose()
        {
            Dispose(true);
        }

        protected void Dispose(bool disposing)
        {
            if (disposing && (_TempFilePath != null) && (File.Exists(_TempFilePath)))
            {
                try
                {
                    File.Delete(_TempFilePath);
                }
                catch (IOException ex)
                {
                    // do nothing
                }
            }
        }

    }
}
