// BenLincoln.DigitalAudio
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of BenLincoln.DigitalAudio.

// BenLincoln.DigitalAudio is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DigitalAudio is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DigitalAudio (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using BLDA = BenLincoln.DigitalAudio;
using BLS = BenLincoln.Shared;

namespace BenLincoln.DigitalAudio
{
    public partial class WaveformDisplay : System.Windows.Forms.Panel, BLS.IProgress
    {
        // colours
        protected Color _BackgroundColour;
        protected Color _WaveformColour;
        protected Color _WaveformMedianColour;
        protected Color _GridColour;
        protected Color _XAxisColour;
        protected Color _CursorColour;

        // grid line counts
        protected int _NumGridLinesX;
        protected int _NumGridLinesY;

        // cursor position (0 - 1)
        protected float _CursorPosition;

        // bitmap cache for wave display
        protected Bitmap _WaveformImage;

        // stream to read the waveform data from on regeneration of the data array
        protected string _WaveformDataStreamPath;

        // sample rate of stream
        protected int _WaveformDataStreamSampleRate;

        // bit depth of stream
        protected int _WaveformDataStreamBitDepth;

        // offset at which the wave data starts in the stream
        protected long _WaveformDataStreamOffset;

        // number of samples in the stream
        protected long _WaveformDataStreamLength;

        // array of 16-bit signed ints to represent the waveform data
        // first index is the sample number
        // second is 0 for min, 1 for media, 2 for max value at that sample
        protected short[][] _WaveformData;

        // maximum oversampling factor for each datapoint of the wave data
        protected int _OversamplingRate;

        // width of the wave data array
        protected int _WaveDataArrayWidth = 16384;

        // waveform drawing enabled
        protected bool _DrawWaveform = true;

        // waveform median -> min/max gradient draw mode enabled
        protected bool _UseGradientDrawMode = true;

        // the waveform data is currently being loaded from the temp file (for loading progress)
        protected bool _IsLoadingData = false;

        // the waveform is currently being drawn (to avoid collisions)
        protected bool _IsDrawing = false;

        // picturebox to display image
        protected PictureBox imgDisplay;

        // image drawing progress
        protected double _Progress;

        // thread for use during complete redraws
        protected Thread _DrawThread;

        // timer to repeatedly invalidate the control during loads
        protected System.Windows.Forms.Timer _Timer;

        // last click position (for seeking)
        protected Point _LastClickPosition;

        // events
        public delegate void WaveformClickedDelegate(object sender, EventArgs e);
        public event WaveformClickedDelegate WaveformClicked;

        #region Properties

        public Color BackgroundColour
        {
            get
            {
                return _BackgroundColour;
            }
            set
            {
                _BackgroundColour = value;
            }
        }

        public Color WaveformColour
        {
            get
            {
                return _WaveformColour;
            }
            set
            {
                _WaveformColour = value;
            }
        }

        public Color WaveformMedianColour
        {
            get
            {
                return _WaveformMedianColour;
            }
            set
            {
                _WaveformMedianColour = value;
            }
        }

        public Color GridColour
        {
            get
            {
                return _GridColour;
            }
            set
            {
                _GridColour = value;
            }
        }

        public Color XAxisColour
        {
            get
            {
                return _XAxisColour;
            }
            set
            {
                _XAxisColour = value;
            }
        }

        public Color CursorColour
        {
            get
            {
                return _CursorColour;
            }
            set
            {
                _CursorColour = value;
            }
        }

        public int NumGridLinesX
        {
            get
            {
                return _NumGridLinesX;
            }
            set
            {
                _NumGridLinesX = value;
            }
        }

        public int NumGridLinesY
        {
            get
            {
                return _NumGridLinesY;
            }
            set
            {
                _NumGridLinesY = value;
            }
        }

        public float CursorPosition
        {
            get
            {
                return _CursorPosition;
            }
            set
            {
                _CursorPosition = value;
            }
        }

        public int OversamplingRate
        {
            get
            {
                return _OversamplingRate;
            }
            set
            {
                _OversamplingRate = value;
            }
        }

        public bool DrawWaveform
        {
            get
            {
                return _DrawWaveform;
            }
            set
            {
                _DrawWaveform = value;
            }
        }

        public bool UseGradientDrawMode
        {
            get
            {
                return _UseGradientDrawMode;
            }
            set
            {
                _UseGradientDrawMode = value;
            }
        }

        public bool IsLoadingData
        {
            get
            {
                return _IsLoadingData;
            }
        }

        public bool IsDrawing
        {
            get
            {
                return _IsDrawing;
            }
        }

        public Point LastClickPosition
        {
            get
            {
                return _LastClickPosition;
            }
        }

        #endregion

        public WaveformDisplay()
            : this(Color.Black, Color.Green, Color.Chartreuse, Color.LightGray, Color.White, Color.Yellow, 9, 10)
        {
        }

        public WaveformDisplay(Color backgroundColour, Color waveformColour, Color waveformMedianColour,
            Color gridColour, Color xAxisColour, Color cursorColour, 
            int numGridLinesX, int numGridLinesY)
        {
            InitializeComponent();
            _BackgroundColour = backgroundColour;
            _WaveformColour = waveformColour;
            _WaveformMedianColour = waveformMedianColour;
            _GridColour = gridColour;
            _XAxisColour = xAxisColour;
            _CursorColour = cursorColour;
            _NumGridLinesX = numGridLinesX;
            _NumGridLinesY = numGridLinesY;
            _CursorPosition = 0;
            _OversamplingRate = 32;
            _WaveformImage = new Bitmap(this.Width, this.Height);
            _LastClickPosition = new Point(0, 0);
            imgDisplay = new PictureBox();
            imgDisplay.Dock = DockStyle.Fill;
            imgDisplay.SizeMode = PictureBoxSizeMode.CenterImage;
            this.Controls.Add(imgDisplay);

            // bind handlers for all the events related to clicking in the display
            MouseDown += new MouseEventHandler(Waveform_MouseDown);
            Click += new EventHandler(ControlClicked);
            imgDisplay.MouseDown += new MouseEventHandler(Waveform_MouseDown);
            imgDisplay.Click += new EventHandler(ControlClicked);
        }

        void Waveform_MouseDown(object sender, MouseEventArgs e)
        {
            _LastClickPosition = new Point(e.X, e.Y);
        }

        protected void ControlClicked(object sender, EventArgs e)
        {
            WaveformClicked(this, e);
        }

        public void SetStream(string inStreamPath, int sampleRate, int bitDepth, long startOffset, long numSamplesInStream)
        {
            if ((bitDepth != 8) && (bitDepth != 16))
            {
                throw new BLS.InvalidParameterValueException("The bit depth of the input stream must be 8 or 16");
            }
            _WaveformDataStreamPath = inStreamPath;
            _WaveformDataStreamSampleRate = sampleRate;
            _WaveformDataStreamBitDepth = bitDepth;
            _WaveformDataStreamOffset = startOffset;
            _WaveformDataStreamLength = numSamplesInStream;
            //_WaveformImage = CreateWaveformImage();
            CreateWaveformImageThreaded();
        }

        protected void CreateWaveformImageThreaded()
        {
            // abort the thread if it's already running - the control is being resized too quickly to keep up
            if ((_DrawThread != null) && (_DrawThread.IsAlive))
            {
                _DrawThread.Abort();
            }
            _DrawThread = new Thread(CreateWaveformImage);
            _DrawThread.Start();
            _Timer = new System.Windows.Forms.Timer();
            _Timer.Tick += new EventHandler(_Timer_Tick);
            _Timer.Start();
        }

        void _Timer_Tick(object sender, EventArgs e)
        {
            if (!_DrawThread.IsAlive)
            {
                _Timer.Stop();
            }
            Invalidate();
        }

        public void RedrawWaveformImage()
        {
            CreateWaveformImageThreaded();
        }

        protected void CreateWaveformImage()
        {
            _IsDrawing = true;
            bool drawWave = _DrawWaveform;
            if (_WaveformDataStreamPath == null)
            {
                drawWave = false;
            }
            else if (_WaveformData == null)
            {
                // regenerate the internal array at a resolution equal to the oversampling rate
                _WaveformData = BuildWaveformArrayFromStream();
            }

            Bitmap waveformImage = new Bitmap(this.Width, this.Height);
            Graphics g = Graphics.FromImage(waveformImage);

            DrawGrid(g);

            // if there is no waveform to draw or drawing is disabled, write text to indicate this
            // otherwise overlay waveform image
            if (!_DrawWaveform)
            {
                DisplayMessage(g, "Waveform display is disabled");
            }
            else if (!drawWave)
            {
                DisplayMessage(g, "No waveform data to display");
            }
            else
            {
                DrawWaveformImage(g);
            }

            _IsDrawing = false;
            _WaveformImage = waveformImage;
        }

        protected void DrawGrid(Graphics g)
        {
            Clear(g, _BackgroundColour);

            // draw background grid, then X axis
            Pen gridPen = new Pen(_GridColour);

            // plus 1 to make calculations easier
            // the values passed as args is actually the number of rows and columns
            int numX = _NumGridLinesX + 1;
            int numY = _NumGridLinesY + 1;

            // draw X-axis lines
            float xLineSpacing = ((float)this.Height / (float)numX);
            for (int lineNum = 0; lineNum < numX; lineNum++)
            {
                Point start = new Point(0, (int)((float)lineNum * xLineSpacing));
                Point end = new Point(this.Width, start.Y);
                g.DrawLine(gridPen, start, end);
            }
            // draw Y-axis lines
            float yLineSpacing = ((float)this.Width / (float)numY);
            for (int lineNum = 0; lineNum < numY; lineNum++)
            {
                Point start = new Point((int)((float)lineNum * yLineSpacing), 0);
                Point end = new Point(start.X, this.Width);
                g.DrawLine(gridPen, start, end);
            }
        }

        //protected void DrawWaveformImage(Graphics g)
        //{
        //    // draw the X-axis zero crossing centerline
        //    Pen gridPen = new Pen(_XAxisColour, 2);
        //    Point start = new Point(0, (this.Height / 2));
        //    Point end = new Point(this.Width, start.Y);
        //    g.DrawLine(gridPen, start, end);

        //    Pen minMaxPen = new Pen(_WaveformColour);
        //    Pen medianPen = new Pen(_WaveformMedianColour);
        //    Pen medianOutlinePen = new Pen(_WaveformColour, 3);

        //    // set the initial point to the zero-crossing at x=0
        //    Point lastMedian = new Point(0, this.Height / 2);

        //    // determine the vertical scaling factor
        //    int zeroCrossing = this.Height / 2;
        //    float vertScale = (float)zeroCrossing / (float)32768;

        //    int arrayPosition = 0;
        //    float arrayStep = (float)this.Width / (float)_WaveDataArrayWidth;
        //    float x = 0;
        //    do
        //    {
        //        int scaledMin = zeroCrossing - (int)((float)_WaveformData[arrayPosition][0] * vertScale);
        //        int scaledMedian = zeroCrossing - (int)((float)_WaveformData[arrayPosition][1] * vertScale);
        //        int scaledMax = zeroCrossing - (int)((float)_WaveformData[arrayPosition][2] * vertScale);
        //        Point min = new Point((int)x, scaledMin);
        //        Point median = new Point((int)x, scaledMedian);
        //        Point max = new Point((int)x, scaledMax);

        //        // should a oversampled vertical line be drawn?
        //        bool drawVerticalLine = true;

        //        if ((scaledMedian == scaledMin) && (scaledMedian == scaledMax))
        //        {
        //            drawVerticalLine = false;
        //        }

        //        if (drawVerticalLine)
        //        {
        //            // connect the last point with the current one
        //            g.DrawLine(medianPen, lastMedian, median);

        //            if (_UseGradientDrawMode)
        //            {
        //                // only use gradients if the min and max aren't the same as the median
        //                if (max.Y != median.Y)
        //                {
        //                    LinearGradientBrush gradientBrush = new LinearGradientBrush(
        //                        median, max, _WaveformMedianColour, _WaveformColour
        //                        );
        //                    Pen gradientPen = new Pen(gradientBrush);
        //                    g.DrawLine(gradientPen, median, max);
        //                }

        //                if (min.Y != median.Y)
        //                {
        //                    LinearGradientBrush gradientBrush = new LinearGradientBrush(
        //                    median, min, _WaveformMedianColour, _WaveformColour
        //                    );
        //                    Pen gradientPen = new Pen(gradientBrush);
        //                    g.DrawLine(gradientPen, median, min);
        //                }
        //            }
        //            else
        //            {
        //                // draw min -> max
        //                g.DrawLine(minMaxPen, min, max);
        //            }
        //        }
        //        else
        //        {
        //            // connect the last point with the current one using a nicer line
        //            g.DrawLine(medianOutlinePen, lastMedian, median);
        //            g.DrawLine(medianPen, lastMedian, median);
        //        }
        //        // set last median to the current median
        //        lastMedian = new Point(median.X, median.Y);

        //        arrayPosition++;
        //        x += arrayStep;
        //    } while (x < this.Width);
        //}

        protected void DrawWaveformImage(Graphics g)
        {
            // draw the X-axis zero crossing centerline
            Pen gridPen = new Pen(_XAxisColour, 2);
            Point start = new Point(0, (this.Height / 2));
            Point end = new Point(this.Width, start.Y);
            g.DrawLine(gridPen, start, end);

            Pen minMaxPen = new Pen(_WaveformColour);
            Pen medianPen = new Pen(_WaveformMedianColour);
            Pen medianOutlinePen = new Pen(_WaveformColour, 3);

            // set the initial point to the zero-crossing at x=0
            Point lastMedian = new Point(0, this.Height / 2);

            // determine the vertical scaling factor
            int zeroCrossing = this.Height / 2;
            float vertScale = (float)zeroCrossing / (float)32768;

            _Progress = 0;
            float arrayPosition = 0;
            int numSamplesPerDataPoint = (int)(_WaveDataArrayWidth / this.Width);
            //float sampleReadStep = (float)_WaveformDataStreamLength / ((float)(arraySize) * (float)numSamplesPerDataPoint);
            //float arrayStep = ((float)_WaveDataArrayWidth / (float)numSamplesPerDataPoint) / (float)this.Width;
            float arrayStep = (float)_WaveDataArrayWidth / ((float)this.Width * (float)numSamplesPerDataPoint);
            for (int x = 0; x < this.Width; x++)
            {
                _Progress = (double)x / (double)this.Width;

                ArrayList minList = new ArrayList();
                ArrayList medList = new ArrayList();
                ArrayList maxList = new ArrayList();
                for (int subSampleNum = 0; subSampleNum < numSamplesPerDataPoint; subSampleNum++)
                {
                    // to avoid going over the array size due to rounding errors
                    int arPos = Math.Min(_WaveDataArrayWidth - 1, (int)arrayPosition);
                    short minSample = _WaveformData[arPos][0];
                    minList.Add(minSample);
                    short medSample = _WaveformData[arPos][1];
                    medList.Add(medSample);
                    short maxSample = _WaveformData[arPos][2];
                    maxList.Add(maxSample);
                    arrayPosition += arrayStep;
                }
                minList.Sort();
                medList.Sort();
                maxList.Sort();

                int scaledMin = zeroCrossing - (int)((float)((short)minList[0]) * vertScale);
                int scaledMedian = zeroCrossing - (int)((float)((short)medList[(medList.Count - 1) / 2]) * vertScale);
                int scaledMax = zeroCrossing - (int)((float)((short)maxList[maxList.Count - 1]) * vertScale);
                Point min = new Point(x, scaledMin);
                Point median = new Point(x, scaledMedian);
                Point max = new Point(x, scaledMax);

                // should a oversampled vertical line be drawn?
                bool drawVerticalLine = true;

                if ((scaledMedian == scaledMin) && (scaledMedian == scaledMax))
                {
                    drawVerticalLine = false;
                }

                if (drawVerticalLine)
                {
                    // connect the last point with the current one
                    g.DrawLine(medianPen, lastMedian, median);

                    if (_UseGradientDrawMode)
                    {
                        // only use gradients if the min and max aren't the same as the median
                        if (max.Y != median.Y)
                        {
                            LinearGradientBrush gradientBrush = new LinearGradientBrush(
                                median, max, _WaveformMedianColour, _WaveformColour
                                );
                            Pen gradientPen = new Pen(gradientBrush);
                            g.DrawLine(gradientPen, median, max);
                        }

                        if (min.Y != median.Y)
                        {
                            LinearGradientBrush gradientBrush = new LinearGradientBrush(
                            median, min, _WaveformMedianColour, _WaveformColour
                            );
                            Pen gradientPen = new Pen(gradientBrush);
                            g.DrawLine(gradientPen, median, min);
                        }
                    }
                    else
                    {
                        // draw min -> max
                        g.DrawLine(minMaxPen, min, max);
                    }
                }
                else
                {
                    // connect the last point with the current one using a nicer line
                    g.DrawLine(medianOutlinePen, lastMedian, median);
                    g.DrawLine(medianPen, lastMedian, median);
                }
                // set last median to the current median
                lastMedian = new Point(median.X, median.Y);

            }
            _Progress = 1;
        }

        protected short[][] BuildWaveformArrayFromStream()
        {
            _IsLoadingData = true;
            int arraySize = _WaveDataArrayWidth;
            short[][] waveformData = new short[arraySize][];
            for (int i = 0; i < arraySize; i++)
            {
                waveformData[i] = new short[3];
            }

            FileStream iStream = new FileStream(_WaveformDataStreamPath, FileMode.Open, FileAccess.Read);

            iStream.Seek(_WaveformDataStreamOffset, SeekOrigin.Begin);
            BinaryReader iReader = new BinaryReader(iStream);

            // read the datapoints
            // if there is more than one sample per datapoint, put the min, median, and max into the array
            // otherwise, read each sample one or more times and use the same value for all three
            double progressFactor = (double)1 / (double)arraySize;
            long streamLength = _WaveformDataStreamLength / (_WaveformDataStreamBitDepth / 8);
            if (arraySize < streamLength)
            {
                int numSamplesPerDataPoint = (int)(streamLength / arraySize);
                if (numSamplesPerDataPoint > _OversamplingRate)
                {
                    numSamplesPerDataPoint = _OversamplingRate;
                }
                float sampleReadStep = (float)streamLength / ((float)(arraySize) * ((float)streamLength / (float)arraySize));
                float fNumSamplesPerDataPoint = ((float)streamLength / (float)arraySize);
                float streamPos = 0;

                for (int dataPointNum = 0; dataPointNum < arraySize; dataPointNum++)
                {
                    streamPos = (float)dataPointNum * fNumSamplesPerDataPoint;
                    _Progress = (progressFactor * (double)dataPointNum);
                    ArrayList dataPoint = new ArrayList();
                    for (int subSampleNum = 0; subSampleNum < numSamplesPerDataPoint; subSampleNum++)
                    {
                        short subSample = 0;
                        if (_WaveformDataStreamBitDepth == 16)
                        {
                            subSample = iReader.ReadInt16();
                            streamPos = Math.Min(_WaveformDataStreamLength - 2, streamPos + sampleReadStep);
                            int shortPos = (int)(streamPos * 2);
                            if (shortPos % 2 == 1)
                            {
                                shortPos++;
                            }
                            iStream.Seek((long)(shortPos * 2), SeekOrigin.Begin);
                        }
                        else
                        {
                            // upconvert from 8-bit to 16-bit
                            subSample = (short)((iReader.ReadByte() << 8) - 32768);
                            streamPos = Math.Min(_WaveformDataStreamLength - 1, streamPos + sampleReadStep);
                            iStream.Seek((long)streamPos, SeekOrigin.Begin);
                        }
                        dataPoint.Add(subSample);
                    }
                    dataPoint.Sort();
                    waveformData[dataPointNum][0] = (short)dataPoint[0];
                    waveformData[dataPointNum][1] = (short)dataPoint[dataPoint.Count / 2];
                    waveformData[dataPointNum][2] = (short)dataPoint[dataPoint.Count - 1];
                }
            }
            else
            {
                float numDataPointsPerSample = (float)_WaveformDataStreamLength / (float)arraySize;
                float currentSampleNum = 0;
                short currentSample = 0;
                // read the first sample
                if (_WaveformDataStreamBitDepth == 16)
                {
                    currentSample = iReader.ReadInt16();
                }
                else
                {
                    // upconvert from 8-bit to 16-bit
                    currentSample = (short)((iReader.ReadByte() << 8) - 32768);
                }
                for (int dataPointNum = 0; dataPointNum < arraySize; dataPointNum++)
                {
                    _Progress = (progressFactor * (double)dataPointNum);
                    if (currentSampleNum > 1)
                    {
                        if (_WaveformDataStreamBitDepth == 16)
                        {
                            currentSample = iReader.ReadInt16();
                        }
                        else
                        {
                            // upconvert from 8-bit to 16-bit
                            currentSample = (short)((iReader.ReadByte() << 8) - 32768);
                        }
                        currentSampleNum -= 1;
                    }
                    waveformData[dataPointNum][0] = (short)currentSample;
                    waveformData[dataPointNum][1] = (short)currentSample;
                    waveformData[dataPointNum][2] = (short)currentSample;
                    currentSampleNum += numDataPointsPerSample;
                }
            }

            iReader.Close();

            _IsLoadingData = false;

            return waveformData;
        }

        //protected short[][] BuildWaveformArrayFromStream(int arraySize)
        //{
        //    short[][] waveformData = new short[arraySize][];
        //    for (int i = 0; i < arraySize; i++)
        //    {
        //        waveformData[i] = new short[3];
        //    }

        //    FileStream iStream = new FileStream(_WaveformDataStreamPath, FileMode.Open, FileAccess.Read);

        //    iStream.Seek(_WaveformDataStreamOffset, SeekOrigin.Begin);
        //    BinaryReader iReader = new BinaryReader(iStream);
           
        //    // read the datapoints
        //    // if there is more than one sample per datapoint, put the min, median, and max into the array
        //    // otherwise, read each sample one or more times and use the same value for all three
        //    double progressFactor = (double)1 / (double)arraySize; 
        //    if (arraySize < _WaveformDataStreamLength)
        //    {
        //        int numSamplesPerDataPoint = (int)(_WaveformDataStreamLength / arraySize);
        //        int sampleReadStep = 1;
        //        if (numSamplesPerDataPoint > _OversamplingRate)
        //        {
        //            sampleReadStep = (int)(numSamplesPerDataPoint / _OversamplingRate);
        //            numSamplesPerDataPoint = _OversamplingRate;
        //        }

        //        for (int dataPointNum = 0; dataPointNum < arraySize; dataPointNum++)
        //        {
        //            _Progress = (progressFactor * (double)dataPointNum);
        //            ArrayList dataPoint = new ArrayList();
        //            for (int subSampleNum = 0; subSampleNum < numSamplesPerDataPoint; subSampleNum++)
        //            {
        //                short subSample = 0;
        //                if (_WaveformDataStreamBitDepth == 16)
        //                {
        //                    subSample = iReader.ReadInt16();
        //                    iStream.Seek((long)((sampleReadStep - 1) * 2), SeekOrigin.Current);
        //                }
        //                else
        //                {
        //                    // upconvert from 8-bit to 16-bit
        //                    subSample = (short)((iReader.ReadByte() << 8) - 32768);
        //                    iStream.Seek((long)(sampleReadStep - 1), SeekOrigin.Current);
        //                }
        //                dataPoint.Add(subSample);
        //            }
        //            dataPoint.Sort();
        //            waveformData[dataPointNum][0] = (short)dataPoint[0];
        //            waveformData[dataPointNum][1] = (short)dataPoint[dataPoint.Count / 2];
        //            waveformData[dataPointNum][2] = (short)dataPoint[dataPoint.Count - 1];
        //        }
        //    }
        //    else
        //    {
        //        float numDataPointsPerSample = (float)_WaveformDataStreamLength / (float)arraySize;
        //        float currentSampleNum = 0;
        //        short currentSample = 0;
        //        // read the first sample
        //        if (_WaveformDataStreamBitDepth == 16)
        //        {
        //            currentSample = iReader.ReadInt16();
        //        }
        //        else
        //        {
        //            // upconvert from 8-bit to 16-bit
        //            currentSample = (short)((iReader.ReadByte() << 8) - 32768);
        //        }
        //        for (int dataPointNum = 0; dataPointNum < arraySize; dataPointNum++)
        //        {
        //            _Progress = (progressFactor * (double)dataPointNum);
        //            if (currentSampleNum > 1)
        //            {
        //                if (_WaveformDataStreamBitDepth == 16)
        //                {
        //                    currentSample = iReader.ReadInt16();
        //                }
        //                else
        //                {
        //                    // upconvert from 8-bit to 16-bit
        //                    currentSample = (short)((iReader.ReadByte() << 8) - 32768);
        //                }
        //                currentSampleNum -= 1;
        //            }
        //            waveformData[dataPointNum][0] = (short)currentSample;
        //            waveformData[dataPointNum][1] = (short)currentSample;
        //            waveformData[dataPointNum][2] = (short)currentSample;
        //            currentSampleNum += numDataPointsPerSample;
        //        }
        //    }

        //    iReader.Close();

        //    return waveformData;
        //}

        protected override void OnPaint(PaintEventArgs pe)
        {
            Bitmap tempBitmap;
            Graphics g;

            if (_IsLoadingData)
            {
                tempBitmap = new Bitmap(this.Width, this.Height);
                g = Graphics.FromImage(tempBitmap);
                DrawGrid(g);
                int percent = (int)((double)100 * _Progress);
                DisplayMessage(g, "Reading Waveform Data: " + percent + "%");
            }
            else if (_IsDrawing)
            {
                tempBitmap = new Bitmap(this.Width, this.Height);
                g = Graphics.FromImage(tempBitmap); 
                DrawGrid(g);
                int percent = (int)((double)100 * _Progress);
                DisplayMessage(g, "Drawing Waveform: " + percent + "%");
            }
            else
            {
                tempBitmap = new Bitmap(_WaveformImage);
                g = Graphics.FromImage(tempBitmap);
                //g.DrawImage(_WaveformImage, 0, 0);
                // draw position cursor if there is a file loaded
                if (_WaveformDataStreamPath != null)
                {
                    Pen gridPen = new Pen(_CursorColour, 3);
                    int cursorPosition = (int)((float)this.Width * _CursorPosition);
                    Point startPoint = new Point(cursorPosition, 0);
                    Point endPoint = new Point(cursorPosition, this.Height);
                    g.DrawLine(gridPen, startPoint, endPoint);
                }
            }

            imgDisplay.Image = tempBitmap;
        }

        protected void Clear(Graphics g, System.Drawing.Color backgroundColour)
        {
            // clear the background
            Rectangle drawArea = new Rectangle(0, 0, this.Width, this.Height);
            SolidBrush bgBrush = new SolidBrush(backgroundColour);
            g.FillRectangle(bgBrush, drawArea);
        }

        protected void DisplayMessage(Graphics g, string text)
        {
            // figure out where the text should be
            SizeF messageSize = g.MeasureString(text, Font);
            int startX = (this.Width - (int)messageSize.Width) / 2;
            int startY = ((this.Height - (int)messageSize.Height) / 2);

            SolidBrush messageBrush = new SolidBrush(_WaveformColour);
            g.DrawString(text, Font, messageBrush, startX, startY);
        }

        protected override void OnResize(EventArgs eventargs)
        {
            base.OnResize(eventargs);
            CreateWaveformImageThreaded();
            //CreateWaveformImage();
            Invalidate();
        }

        public double Progress()
        {
            return _Progress;
        }

    }
}
