// Blood Omen 2 Control Station
// Copyright 2007-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of Blood Omen 2 Control Station.

// Blood Omen 2 Control Station is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// Blood Omen 2 Control Station is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Blood Omen 2 Control Station (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32;
using BenLincoln.TheLostWorlds.CDGameDotErg;
using BLGDE = BenLincoln.TheLostWorlds.CDGameDotErg;

namespace BloodOmen2ControlStation
{
    public partial class frmMain : Form
    {
        protected string _OutputPath;
        protected bool _FileExisted;
        protected BLGDE.GameLevelCollection _Levels;
        protected BLGDE.ERGVariableCollection _Variables;

        public frmMain()
        {
            InitializeComponent();
            SetVersionDisplay();
            _Levels = CreateLevels();
            PopulateLevelDropDown();
            _Variables = InitializeVariables();
            _OutputPath = GetBO2InstallPath();
            if (_OutputPath == "")
            {
                _OutputPath = Path.GetDirectoryName(Application.ExecutablePath);
                MessageBox.Show("Blood Omen 2 Control Station was unable to locate the Blood Omen 2 installation folder on this computer. " +
                    "The game.erg file will be output to the location of the Blood Omen 2 Control Station program instead (" + 
                    _OutputPath + ").",
                    "InstallPath Registry Key is Missing", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            _OutputPath += @"\game.erg";
            _FileExisted = false;
            if (File.Exists(_OutputPath))
            {
                LoadSettingsFromExistingFile();
                _FileExisted = true;
            }
            BindVariablesToControls();
        }

        protected string GetBO2InstallPath()
        {
            RegistryKey regKey;
            regKey = Registry.LocalMachine;
            try
            {
                regKey = regKey.OpenSubKey(@"SOFTWARE\Crystal Dynamics\Blood Omen 2");
                string installPath = (string)regKey.GetValue("InstallPath");
                if (Directory.Exists(installPath + @"\data"))
                {
                    return installPath + @"\data";
                }
            }
            catch (Exception ex)
            {
                return "";
            }
            return "";
        }

        protected BLGDE.GameLevelCollection CreateLevels()
        {
            BLGDE.GameLevelCollection levels = new BLGDE.GameLevelCollection(1);
            try
            {
                Assembly asm = Assembly.GetExecutingAssembly();
                Stream inStream = asm.GetManifestResourceStream("BloodOmen2ControlStation.Resources." + "levellist.txt");
                StreamReader inReader = new StreamReader(inStream);
                string levelList = inReader.ReadToEnd();
                inReader.Close();
                inStream.Close();
                string[] lines = levelList.Split(new string[] { "\r\n" }, StringSplitOptions.None);
                levels = new BLGDE.GameLevelCollection(lines.GetUpperBound(0) + 1);
                int currentLevel = 0;
                foreach (string line in lines)
                {
                    if (line.Trim() != "")
                    {
                        string[] levelInfo = line.Split('\t');
                        inStream = asm.GetManifestResourceStream("BloodOmen2ControlStation.Resources." + levelInfo[0]);
                        levels.Levels[currentLevel] = LoadLevelFromStream(levelInfo[1], inStream);
                        inStream.Close();
                        currentLevel++;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Blood Omen 2 Control Station was unable to load embedded level information:" + Environment.NewLine +
                    ex.Message + Environment.NewLine +
                    "This is a fatal error and the application must exit now.",
                    "Error Loading Embedded Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
                System.Environment.ExitCode = 1;
                Application.Exit();
            }
            return levels;
        }

        protected BLGDE.GameLevel LoadLevelFromStream(string displayName, Stream inStream)
        {
            BLGDE.GameLevel level = new BLGDE.GameLevel(displayName);
            StreamReader inReader = new StreamReader(inStream);
            string rawLevelData = inReader.ReadToEnd();
            string[] lines = rawLevelData.Split(new string[] { "\r\n" }, StringSplitOptions.None);
            level.Areas = new GameArea[lines.GetUpperBound(0) + 1];
            int areaNum = 0;
            foreach (string line in lines)
            {
                if (line.Trim() != "")
                {
                    string[] areaInfo = line.Split('\t');
                    level.Areas[areaNum] = new BLGDE.GameArea(areaInfo[1], areaInfo[2], areaInfo[0], level);
                    areaNum++;
                }
            }
            return level;
        }

        protected BLGDE.ERGVariableCollection InitializeVariables()
        {
            BLGDE.ERGVariableCollection vars = new ERGVariableCollection(1);
            try
            {
                Assembly asm = Assembly.GetExecutingAssembly();
                Stream inStream = asm.GetManifestResourceStream("BloodOmen2ControlStation.Resources." + "vars.txt");
                StreamReader inReader = new StreamReader(inStream);
                string varList = inReader.ReadToEnd();
                inReader.Close();
                inStream.Close();
                string[] lines = varList.Split(new string[] { "\r\n" }, StringSplitOptions.None);
                vars = new BLGDE.ERGVariableCollection(lines.GetUpperBound(0));
                int currentVar = 0;
                foreach (string line in lines)
                {
                    if (line.Trim() != "")
                    {
                        string[] varInfo = line.Split('\t');
                        string varCategory = varInfo[0];
                        string varName = varInfo[1];
                        string varType = varInfo[2].ToLower();
                        string varDefault = varInfo[3];
                        switch (varType)
                        {
                            case "bool":
                                vars.Variables[currentVar] = new BLGDE.ERGVariableBool(
                                    varCategory, varName, null, BLGDE.ERGVariableBool.StringToBool(varDefault), false);
                                break;
                            case "invbool":
                                vars.Variables[currentVar] = new BLGDE.ERGVariableBool(
                                    varCategory, varName, null, BLGDE.ERGVariableBool.StringToBool(varDefault), true);
                                break;
                            case "int":
                                vars.Variables[currentVar] = new BLGDE.ERGVariableInt(
                                    varCategory, varName, null, BLGDE.ERGVariableInt.StringToInt(varDefault));
                                break;
                            case "string":
                                vars.Variables[currentVar] = new BLGDE.ERGVariableString(
                                    varCategory, varName, null, varDefault.Trim());
                                break;
                            case "startarea":
                                vars.Variables[currentVar] = new BLGDE.ERGVariableStartArea(
                                    varCategory, varName, null, null, null,
                                    varDefault, _Levels);
                                break;
                        }
                        currentVar++;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Blood Omen 2 Control Station was unable to load embedded variable information:" + Environment.NewLine +
                    ex.Message + Environment.NewLine +
                    "This is a fatal error and the application must exit now.",
                    "Error Loading Embedded Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
                System.Environment.ExitCode = 1;
                Application.Exit();
            }
            return vars;
        }

        protected void PopulateLevelDropDown()
        {
            cbLevel.Items.Clear();
            foreach (BLGDE.GameLevel level in _Levels.Levels)
            {
                cbLevel.Items.Add(level.Name);
            }
        }

        private void cbLevel_SelectedIndexChanged(object sender, EventArgs e)
        {
            PopulateAreaDropDown();
        }

        protected void PopulateAreaDropDown()
        {
            string levelName = cbLevel.SelectedItem.ToString();
            BLGDE.GameLevel level = _Levels.GetLevelByName(levelName);
            cbArea.Items.Clear();
            foreach (BLGDE.GameArea area in level.Areas)
            {
                cbArea.Items.Add(area.Name);
            }
            cbArea.SelectedIndex = 0;
        }

        protected void LoadSettingsFromExistingFile()
        {
            try
            {
                FileStream inStream = new FileStream(_OutputPath, FileMode.Open, FileAccess.Read);
                StreamReader inReader = new StreamReader(inStream);
                string existingData = inReader.ReadToEnd();
                inReader.Close();
                inStream.Close();
                string[] lines = existingData.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
                ArrayList unrecognizedVars = new ArrayList();
                foreach (string line in lines)
                {
                    if (line.Trim() != "")
                    {
                        string[] lineData = line.Split('=');
                        string varFullPath = lineData[0];
                        string varValue = lineData[1];
                        // remove unnecessary double-quotes
                        varFullPath = varFullPath.Replace("\"", "");
                        varValue = varValue.Replace("\"", "");
                        string varCategory;
                        string varName;
                        if (varFullPath.Contains("\\"))
                        {
                            string[] varData = varFullPath.Split('\\');
                            varCategory = varData[0];
                            varName = varData[1];
                        }
                        else
                        {
                            varCategory = "";
                            varName = varFullPath;
                        }
                        bool invertValue = false;
                        if (varCategory.StartsWith("-"))
                        {
                            invertValue = true;
                            varCategory = varCategory.Substring(1);
                        }
                        if (_Variables.ContainsVariable(varCategory, varName))
                        {
                            int varNum = _Variables.GetIndexFromName(varCategory, varName);
                            _Variables.Variables[varNum].SetValue(varValue, invertValue);
                        }
                        else
                        {
                            unrecognizedVars.Add("'" + varCategory + "'\\'" + varName + "'");
                        }

                    }
                }
                if (unrecognizedVars.Count > 0)
                {
                    StringBuilder builder = new StringBuilder();
                    foreach (string uVar in unrecognizedVars)
                    {
                        builder.Append(uVar + Environment.NewLine);
                    }
                    MessageBox.Show("The following variables are unrecognized:" +
                       Environment.NewLine + Environment.NewLine + builder.ToString() + Environment.NewLine +
                       "Blood Omen 2 Control Station will ignore these variables and will not write them " +
                       "to game.erg if the file is updated. If you wish to save these variables, please make a " + 
                       "copy of your game.erg before modifying it with this application.",
                       "Unknown Variables", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show("An error occurred while loading data from the file '" + _OutputPath + "':" +
                    Environment.NewLine + ex.Message + Environment.NewLine +
                    "Please make sure the file is accessible and not in use by another program. " +
                    "Default values will be used instead.", "Error Reading Existing File",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        protected void BindVariablesToControls()
        {
            int varIndex;

            try
            {
                varIndex = _Variables.GetIndexFromName("Blood Omen Specific Flags", "dont show monster prints");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkMonsterMessages });

                varIndex = _Variables.GetIndexFromName("Blood Omen Specific Flags", "show detector debug stuff");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkShowDetector });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "Dark gifts always available");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkInfiniteRage });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "hide Kain during dark gift targeting");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkHideKain });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "ignore water for Kain");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkImmuneToWater });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "kain's invulnerable");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkInvulnerable });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show checkpoint info");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkCheckpointInfo });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show elapsed player time");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkElapsedTime });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's animations");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkAnimations });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's behaviors");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkBehaviour });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's charge move points");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkChargeMove });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's hit points");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkHitPoints });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's lore");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkLore });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "show Kain's position");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkShowPosition });

                varIndex = _Variables.GetIndexFromName("BO2 Kain Debug Flags", "weapons are not breakable");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkUnbreakable });

                varIndex = _Variables.GetIndexFromName("BO2 New Feature Enable Flags", "enable new camera");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkNewCamera });

                varIndex = _Variables.GetIndexFromName("BO2 New Feature Enable Flags", "show dark gift name in menu");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkShowDGNames });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Disable script debug print");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkDebuggingMessages });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Display Current Regions");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkCurrentRegions });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Draw UI");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkDrawUI });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Enable creature lighting");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkCreatureLighting });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Show Mem Totals");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkMemTotals });

                varIndex = _Variables.GetIndexFromName("Debug Common", "Show UI debug info");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkAttitudeDot });

                varIndex = _Variables.GetIndexFromName("Debug Engine", "Show Memory Usage");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkShowMemory });

                varIndex = _Variables.GetIndexFromName("", "NoFPS");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkFPS });

                varIndex = _Variables.GetIndexFromName("", "PlayMovie");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkIntroMovies });

                varIndex = _Variables.GetIndexFromName("", "ReadLighting");
                _Variables.Variables[varIndex].BindControls(new Control[] { chkEnvironmentalLighting });

                varIndex = _Variables.GetIndexFromName("", "Level");
                _Variables.Variables[varIndex].BindControls(new Control[] { cbLevel, cbArea, lblDescription });

            }
            catch (KeyNotFoundException kEx)
            {
                MessageBox.Show("Blood Omen 2 Control Station was unable to load embedded variable information:" + Environment.NewLine +
                    kEx.Message + Environment.NewLine +
                    "This is a fatal error and the application must exit now.",
                    "Error Loading Embedded Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
                System.Environment.ExitCode = 1;
                Application.Exit();
            }
        }

        protected void SetVersionDisplay()
        {
            Assembly asm = Assembly.GetExecutingAssembly();
            AssemblyName name = asm.GetName();
            string version = name.Version.Major.ToString() + "." + name.Version.Minor.ToString();
            lblVersion.Text = "Version " + version + ", build " + name.Version.Build.ToString();
        }

        #region File Saving

        protected void Revert()
        {
            string defaults = "";
            try
            {
                Assembly asm = Assembly.GetExecutingAssembly();
                Stream inStream = asm.GetManifestResourceStream("BloodOmen2ControlStation.Resources." + "DefaultERG.txt");
                StreamReader inReader = new StreamReader(inStream);
                defaults = inReader.ReadToEnd();
                inReader.Close();
                inStream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Blood Omen 2 Control Station was unable to load embedded configuration information:" + Environment.NewLine +
                    ex.Message + Environment.NewLine +
                    "This is a fatal error and the application must exit now.",
                    "Error Loading Embedded Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
                System.Environment.ExitCode = 1;
                Application.Exit();
            }
            if (defaults != "")
            {
                WriteFile(defaults);
            }

        }

        protected void Save()
        {
            StringBuilder builder = new StringBuilder();
            foreach (ERGVariable var in _Variables.Variables)
            {
                builder.Append(var.GetFullName() + "=" + var.GetValue() + Environment.NewLine);
            }
            WriteFile(builder.ToString());
        }

        protected void WriteFile(string data)
        {
            try
            {
                this.Cursor = Cursors.WaitCursor;
                FileStream outStream = new FileStream(_OutputPath, FileMode.Create, FileAccess.Write);
                StreamWriter outWriter = new StreamWriter(outStream);
                outWriter.Write(data);
                outWriter.Close();
                outStream.Close();
                this.Cursor = Cursors.Default;
                //string saveType = "updated";
                //if (!_FileExisted)
                //{
                //    saveType = "created";
                //}
                ////MessageBox.Show("File was " + saveType + " successfully.", "Save Complete",
                ////    MessageBoxButtons.OK, MessageBoxIcon.Information);
                //_FileExisted = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show("An error occurred while saving data to the file '" + _OutputPath + "':" +
                    Environment.NewLine + ex.Message + Environment.NewLine +
                    "Please make sure the file is accessible and not in use by another program. ",
                    "Error Writing Output File",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion


        #region Button Commands

        private void btnRevert_Click(object sender, EventArgs e)
        {
            DialogResult result = MessageBox.Show(
                "Do you want to discard all changes and revert to the version of game.erg included with Blood Omen 2?" +
                Environment.NewLine + "This action cannot be undone.", "Revert?", 
                MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            if (result == DialogResult.Yes)
            {
                Revert();
                LoadSettingsFromExistingFile();
            }
        }

        private void btnSave_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        #endregion


    }
}